/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.awt.tate;

import net.hizlab.kagetaka.Resource;
import net.hizlab.kagetaka.awt.ImageCreator;
import net.hizlab.kagetaka.rendering.Option;

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Panel;
import java.awt.Point;
import java.awt.SystemColor;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

/**
 * ɽΥܥǤ
 * 
 * @kagetaka.bugs ľ󲽤ϡꥹʤ¸ʤޤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.3 $
 */
public class TextArea
	extends TextComponent
{
	private static final String RESOURCE    = "net.hizlab.kagetaka.awt.tate.Resource";
	private static final String BASE        = "textarea";
	private static       int    nameCounter = 0;
	
	private static final int SCROLLBAR_SIZE = 16;
	
	/** ξ */
	public static final int SCROLLBARS_BOTH = 0;
	/** ĤΤ */
	public static final int SCROLLBARS_VERTICAL_ONLY = 1;
	/** Τ */
	public static final int SCROLLBARS_HORIZONTAL_ONLY = 2;
	/** ̵ */
	public static final int SCROLLBARS_NONE = 3;
	
	/** @serial СβĻ° */
	private int       scrollbarVisibility;
	/** @serial С */
	private Scrollbar scrollbarHorizontal;
	/** @serial ĥС */
	private Scrollbar scrollbarVertical;
	
	/**
	 * 󥹥󥹤ޤ
	 * 
	 * @param     option ץ
	 * @param     ic     ᡼ꥨ
	 */
	public TextArea(Option option, ImageCreator ic)
	{
		this(option, ic, null, 0, 0, SCROLLBARS_BOTH);
	}
	
	/**
	 * ꤵ줿ƥȤޤ।󥹥󥹤ޤ
	 * 
	 * @param     option  ץ
	 * @param     ic      ᡼ꥨ
	 * @param     text    ƥ
	 */
	public TextArea(Option option, ImageCreator ic, String text)
	{
		this(option, ic, text, 0, 0, SCROLLBARS_BOTH);
	}
	
	/**
	 * ꤵ줿Υ󥹥󥹤ޤ
	 * 
	 * @param     option  ץ
	 * @param     ic      ᡼ꥨ
	 * @param     rows    Կ
	 * @param     columns 
	 */
	public TextArea(Option option, ImageCreator ic, int rows, int columns)
	{
		this(option, ic, null, rows, columns, SCROLLBARS_BOTH);
	}
	
	/**
	 * ꤵ줿ƥȤΥ󥹥󥹤ޤ
	 * 
	 * @param     option  ץ
	 * @param     ic      ᡼ꥨ
	 * @param     text    ƥ
	 * @param     rows    Կ
	 * @param     columns 
	 */
	public TextArea(Option option, ImageCreator ic, String text,
	                int rows, int columns)
	{
		this(option, ic, text, rows, columns, SCROLLBARS_BOTH);
	}
	
	/**
	 * ꤵ줿ƥȤΥ󥹥󥹤ޤ
	 * 
	 * @param     option  ץ
	 * @param     ic      ᡼ꥨ
	 * @param     text    ƥ
	 * @param     rows    Կ
	 * @param     columns 
	 * @param     scrollbars СβĻ°
	 */
	public TextArea(Option option, ImageCreator ic, String text,
	                int rows, int columns, int scrollbars)
	{
		super(option, ic, text, rows, columns, true);
		this.scrollbarVisibility = scrollbars;
		
		if (scrollbars == SCROLLBARS_BOTH ||
		    scrollbars == SCROLLBARS_HORIZONTAL_ONLY) {
			scrollbarHorizontal = new Scrollbar(Scrollbar.HORIZONTAL);
		}
		if (scrollbars == SCROLLBARS_BOTH ||
		    scrollbars == SCROLLBARS_VERTICAL_ONLY  ) {
			scrollbarVertical   = new Scrollbar(Scrollbar.VERTICAL  );
		}
		
		setFrameSize(INSET_V,
		             INSET_H,
		             INSET_V + (scrollbarHorizontal != null ? SCROLLBAR_SIZE : 0),
		             INSET_H + (scrollbarVertical   != null ? SCROLLBAR_SIZE : 0));
	}
	
	/**
	 * ΥƥȥꥢΥѥ᡼ʸ֤ޤ
	 * 
	 * @return    ѥ᡼ʸ
	 */
	protected String paramString()
	{
		String str = super.paramString();
		String sbVisStr;
		switch (scrollbarVisibility) {
		case SCROLLBARS_BOTH:
			sbVisStr = "both";
			break;
		case SCROLLBARS_VERTICAL_ONLY:
			sbVisStr = "vertical-only";
			break;
		case SCROLLBARS_HORIZONTAL_ONLY:
			sbVisStr = "horizontal-only";
			break;
		case SCROLLBARS_NONE:
			sbVisStr = "none";
			break;
		default:
			sbVisStr = "invalid display policy";
		}
		return str + ",rows=" + getRows()
		           + ",columns=" + getColumns()
		           + ", scrollbarVisibility=" + sbVisStr;
	}
	
	/** ݡͥȤΥǥեȤ֤̾ */
	String constructComponentName()
	{
		return BASE + nameCounter++;
	}
	
	/** ߤΥơˤä֤˺ */
	protected void refresh(Graphics g, Dimension size, int state, boolean focus)
	{
		super.refresh(g, size, state, focus);
		
		Point     viewPosition = getViewPosition();
		Dimension textSize     = getTextSize    ();
		Insets    frameSize    = getFrameSize   ();
		if (scrollbarHorizontal != null) {
			int page = size.width  - frameSize.left - frameSize.right;
			scrollbarHorizontal.setBounds(2,
			                              size.height - SCROLLBAR_SIZE - 2,
			                              size.width  - (scrollbarVertical   != null ? SCROLLBAR_SIZE : 0) - 4,
			                              SCROLLBAR_SIZE);
//System.out.println(viewPosition.x+","+page+","+0+","+textSize.width);
			scrollbarHorizontal.setReverseValues (viewPosition.x, page, 0, textSize.width);
			scrollbarHorizontal.setBlockIncrement(Math.max(10, page - 10));
			scrollbarHorizontal.setUnitIncrement (10);
			scrollbarHorizontal.setEnabled       (scrollbarHorizontal.canScroll());
			scrollbarHorizontal.paint(g);
		}
		if (scrollbarVertical != null) {
			int page = size.height - frameSize.top - frameSize.bottom;
			scrollbarVertical  .setBounds(size.width  - SCROLLBAR_SIZE - 2,
			                              2,
			                              SCROLLBAR_SIZE,
			                              size.height - (scrollbarHorizontal != null ? SCROLLBAR_SIZE : 0) - 4);
//System.out.println(viewPosition.y+","+page+","+0+","+textSize.height);
			scrollbarVertical  .setValues        (viewPosition.y, page, 0, textSize.height);
			scrollbarVertical  .setBlockIncrement(Math.max(getLineHeight(), page - getLineHeight()));
			scrollbarVertical  .setUnitIncrement (getLineHeight());
			scrollbarVertical  .setEnabled       (scrollbarVertical.canScroll());
			scrollbarVertical  .paint(g);
		}
		if (scrollbarHorizontal != null && scrollbarVertical != null) {
			g.setColor(SystemColor.control);
			g.fillRect(size.width  - SCROLLBAR_SIZE - 2,
			           size.height - SCROLLBAR_SIZE - 2,
			           SCROLLBAR_SIZE, SCROLLBAR_SIZE);
		}
	}
	
	/** {@inheritDoc} */
	public int getColumns()
	{
		return super.getColumns();
	}
	
	/** {@inheritDoc} */
	public void setColumns(int columns)
	{
		super.setColumns(columns);
	}
	
	/** {@inheritDoc} */
	public int getRows()
	{
		return super.getRows();
	}
	
	/** {@inheritDoc} */
	public void setRows(int rows)
	{
		super.setRows(rows);
	}
	
	/** {@inheritDoc} */
	public boolean getWordwrap()
	{
		return super.getWordwrap();
	}
	
	/** {@inheritDoc} */
	public void setWordwrap(boolean b)
	{
		super.setWordwrap(b);
	}
	
	/** {@inheritDoc} */
	Window showInputBox(Frame owner, int x, int y)
	{
		Point  point  = getLocationOnScreen();
		Input  input  = new Input(owner);
		Insets insets = input.getInsets();
		input.setLocation(point.x + x - insets.left, point.y + y - insets.top);
		input.show();
		input.text.requestFocus();
		input.toFront();
		
		return input;
	}
	
//### Input
	/** ϥ */
	private class Input
		extends Dialog
		implements ActionListener
	{
		private static final String COMMAND_OK     = "ok";
		private static final String COMMAND_CANCEL = "cancel";
		private java.awt.TextArea text;
		private boolean isClosed;
		
		/** 󥹥󥹤 */
		private Input(Frame parent)
		{
			super(parent, Resource.getMessage(RESOURCE, "input.title", null), false);
			
			text = new java.awt.TextArea(value, getRows(), getColumns(), scrollbarVisibility);
			text.setForeground(TextArea.this.getForeground());
			text.setBackground(TextArea.this.getBackground());
			text.setFont      (TextArea.this.getFont      ());
			add(text);
			
			Panel panel;
			add(panel = new Panel(), BorderLayout.SOUTH);
			panel.setLayout(new GridLayout(1, 2, 10, 5));
			
			java.awt.Button button;
			panel.add(button = new java.awt.Button(Resource.getMessage(RESOURCE, "textarea.button.ok"    , null)));
			button.setActionCommand (COMMAND_OK    );
			button.addActionListener(this          );
			panel.add(button = new java.awt.Button(Resource.getMessage(RESOURCE, "textarea.button.cancel", null)));
			button.setActionCommand (COMMAND_CANCEL);
			button.addActionListener(this          );
			
			pack();
			
			// ꥹʤϿ
			addWindowListener(
				new WindowAdapter()
				{
					/** ɥĤ */
					public void windowClosing(WindowEvent e)
					{
						close(false);
					}
					
					/** ƥ֤ǤϤʤʤä */
					public void windowDeactivated(WindowEvent e)
					{
						close(true);
					}
				}
			);
			text.addKeyListener(
				new KeyAdapter()
				{
					/** 줿 */
					public void keyPressed(KeyEvent e)
					{
						switch (e.getKeyCode()) {
						case KeyEvent.VK_ESCAPE:
							e.consume();
							close(false);
							break;
						case KeyEvent.VK_ENTER:
							if ((e.getModifiers() & InputEvent.CTRL_MASK) != 0) {
								e.consume();
								close(true);
							}
							break;
						}
					}
				}
			);
		}
		
		/** Ĥ */
		private void close(boolean commit)
		{
			synchronized (this) {
				if (isClosed)
					return;
				isClosed = true;
			}
			
			if (commit)
				setText(text.getText());
			dispose();
		}
		
		/**  */
		public void update(Graphics g)
		{
			paint(g);
		}
		
		/**  */
		public void actionPerformed(ActionEvent e)
		{
			close((COMMAND_OK.compareTo(e.getActionCommand()) == 0));
		}
	}
	
//### Scrollbar
	/** С */
	private class Scrollbar
		extends InnerScrollbar
	{
		private int orientation;
		
		/** 󥹥󥹤 */
		private Scrollbar(int orientation)
		{
			super(orientation, TextArea.this);
			this.orientation = orientation;
		}
		
		/**  */
		void repaint()
		{
			repaintForce();
		}
		
		/** ͤѹ줿 */
		void changedValue()
		{
			Point viewPosition = getViewPosition();
			if (orientation == Scrollbar.HORIZONTAL) {
				int x = getReverseValue();
				if (x == viewPosition.x)
					return;
				setViewPosition(x, viewPosition.y);
			} else {
				int y = getValue();
				if (y == viewPosition.y)
					return;
				setViewPosition(viewPosition.x, y);
			}
			repaintForce();
		}
	}
}
